package xyz.maxwellj.chookspeak 

import java.io.File

fun splitList(input: String): List<String> {
    val result = mutableListOf<String>()
    var currentToken = StringBuilder()
    var insideQuotes = false
    
    for (char in input) {
        when {
            char == '"' -> {
                currentToken.append(char)
                insideQuotes = !insideQuotes
            }
            char.isWhitespace() && !insideQuotes -> {
                if (currentToken.isNotEmpty()) {
                    result.add(currentToken.toString())
                    currentToken = StringBuilder()
                }
            }
            else -> currentToken.append(char)
        }
    }
    
    if (currentToken.isNotEmpty()) {
        result.add(currentToken.toString())
    }
    
    return result
}

object log {
    var isLogging = 0
    fun log(input: String) {
        if (isLogging == 1) {
            println(input)
        }
    }
    fun error(input: String) {
        println("\u001b[31m$input\u001b[0m")
    }
    fun note(input: String) {
        println("\u001b[32m$input\u001b[0m")
    }
    fun info(input: String) {
        println("\u001b[33m$input\u001b[0m")
    }
}

object files {
    val outputIndexJS = File("out/index.js")
    val outputMessageJS = File("out/message.js}")
    val outputHTML = File("out/index.html") 
    val generatedIndexJS = mutableListOf<String>()
    val generatedMessageJS = mutableListOf<String>()
    val generatedHTML = mutableListOf<String>()
   
    var errorStatus = 0

    fun write() {
        if (errorStatus != 0) {
            log.error("\u001b[1mTranspile aborted due to code errors. If you believe this is a mistake please make a bug report. Run again with verbose mode (-v) and upload the log.")
            return
        }
        log.note("Finished! Writing...")
        outputIndexJS.writeText("")
        outputMessageJS.writeText("")
        outputHTML.writeText("")
 
        for (item in generatedHTML) {
            outputHTML.appendText("$item\n")
        }
        for (item in generatedMessageJS) {
            outputMessageJS.appendText("$item\n")
        }
        for (item in generatedIndexJS) {
            outputIndexJS.appendText("$item\n")
        }
        log.note("All done! Your files are in ./out/")
    }
}

fun testValidity(input: List<String>, num: Int): Int {
    val validOperations = listOf("=", "++")
    val declarations = listOf("val", "var")
    val initialising = declarations.contains(input[0])
    
    log.log("Testing validity of ${input.toString()}...")
    log.log("Testing a newly initialised variable is $initialising")
    
    if (initialising) {
        if (input[2] != "=") {
            log.error("""Error at line $num: Define your variable with "=" not "${input[2]}"""")
            files.errorStatus = 1
            return 1
        }
    } else {
        if (!validOperations.contains(input[1])) {
            log.error("""Error at line $num: Invalid operation "${input[1]}". Valid operations are: ${validOperations.joinToString(", ")}""")
            files.errorStatus = 1
            return 1
        }
    }
    
    return 0
}

fun main(args: Array<String>) {
    log.note("Starting transpiler...")
    try {
        if (args[1] == "-v") {
            log.isLogging = 1
        }
    } catch (e: Exception) {}
    var fileLocation: File
    try {
        fileLocation = File(args[0])
    } catch (e: Exception) {
        log.error("Error: Please specify a file")
        return
    }
    if (!fileLocation.exists()) {
        log.error("Error: Your Chookspeak file does not exist.")
        return
    }
    val file = fileLocation.readLines()
    var lineCounter = 1
    var commenting = 0
    var eggName = ""
    val status = mutableListOf<String>()
    val variables = mutableListOf<String>()
    val keywords = listOf("init", "sendMessage", "html", "log", "js", "val", "var", "}")
    log.note("File loaded!")
    for (line in file) {
        val lineArgsUnfiltered = splitList(line) 
        val lineArgs = lineArgsUnfiltered.filterNot { it == "" }
        if (lineArgs.size == 0) {
            continue
        }
        log.log("line ${lineArgs.toString()}")
        var slc = 0
        for (item in lineArgs) {
            if (item == "//") {
                commenting = 1
            } else if (item == "/*") {
                commenting = 2
                slc = 1
            } else if (item == "*/" && commenting == 2 && slc == 0) {
                commenting = 0
            }
        }
        if (commenting == 0) {
            if (lineCounter == 1) {
                if (lineArgs[0] != "egg") {
                    log.error("Error: Please specify your egg on your first line of code.")
                    return
                } else {
                    eggName = lineArgs[1]
                    files.generatedIndexJS.add("function egg$eggName() {")                
                    files.generatedHTML.add("""<div id="$eggName" style="display: hidden">""")
                }
            }
            if (lineArgs[0] == "}") {
                try {
                    status.removeLast()
                } catch (e: Exception) {}
                log.log("status ${status.toString()}")
            }
            if (lineArgs.size > 1) {
                if (lineArgs[1] == "{") {
                    status.add(lineArgs[0])
                    log.log("status ${status.toString()}")
                }
            }
            if (status == listOf("init")) {
                if (lineArgs[0] == "log") {
                    files.generatedIndexJS.add("console.log(${lineArgs[1]});")
                }
                if (lineArgs[0] == "sendMessage") {
                    files.generatedIndexJS.add("""const egg${eggName}Message = { "type" = "egg-$eggName", "username" = username, "token" = md5(password), "content": ${lineArgs[1]}};${""}""")
                    files.generatedIndexJS.add("ws.send(JSON.stringify(egg${eggName}Message))")
                }
                if (lineArgs[0] == "val") {
                    if (lineArgs[2] == "=") {
                        files.generatedIndexJS.add("const ${lineArgs[1]} = ${lineArgs[3]};")
                    } else {
                        log.error("Error on line $lineCounter: Define your value with '=', not ${lineArgs[2]}")
                        files.errorStatus = 1
                    }
                }
                if (lineArgs[0] == "var") {
                    if (lineArgs[2] == "=") {
                        testValidity(lineArgs, lineCounter)
                        files.generatedIndexJS.add("let ${lineArgs[1]} = ${lineArgs[3]};")
                        variables.add(lineArgs[1])    
                    } else {
                        log.error("Error on line $lineCounter: Define your variable with '=', not ${lineArgs[2]}")
                        files.errorStatus = 1
                    }
                }
                if (variables.contains(lineArgs[0])) {
                    if (testValidity(lineArgs, lineCounter) == 0) {
                        files.generatedIndexJS.add(lineArgs.joinToString(" "))
                    }
                } else if (!keywords.contains(lineArgs[0])){
                    log.info("Info: On line $lineCounter, I don't know how to ${lineArgs[0]} yet. Maybe you forgot to define a variable? Or if the feature isn't implemented yet, check back later.")
                    log.info("I'll skip that line. Your project will still be outputted, however expect bugs.")
                }
            }
            if (status == listOf("init", "html") && lineArgs[0] != "html") {
                files.generatedHTML.add("${lineArgs.joinToString(" ")}")
            }
            if (status == listOf("init", "js") && lineArgs[0] != "js") {
                files.generatedIndexJS.add("${lineArgs.joinToString(" ")}")
            }           
            lineCounter ++
        } else if (commenting == 1) {
            commenting = 0
        } else if (commenting == 2) {
            commenting = 0
        }
    }
    files.generatedIndexJS.add("}")
    files.generatedHTML.add("</div>")
    files.write()
}
